// -*- c++ -*-

#ifndef KERNEL_LOCK_H
#define KERNEL_LOCK_H

#include <flux/x86/eflags.h>
#include <flux/x86/proc_reg.h>
#include <flux/x86/pc/pic.h>

#include "config.h"

// 
// Global kernel lock.  When held, irqs are also disabled on the
// current CPU (preventing nested irq handling, and preventing the
// current thread from being preempted).  It must only be held for
// very short amounts of time.
// 
class kernel_lock_t
{
public:
#ifdef PROFILE
  kernel_lock_t();
#endif

  void lock();			// lock kernel, disable irqs on CPU
  void clear();			// unlock kernel, enable irqs on CPU
  void clear_irqdisable();	// unlock kernel, but disable irqs on CPU
  bool test() const;
  bool test_and_set();

  void set(bool state);

private:
#ifdef PROFILE
  char _mask_master, _mask_slave;
  bool _is_set;
#endif
};

// 
// kernel_lock_t inlines
// 

#ifndef PROFILE			// not profiling

inline void
kernel_lock_t::lock()
{
  cli();
}

inline void
kernel_lock_t::clear()
{
  sti();
}

inline void
kernel_lock_t::clear_irqdisable()
{
  cli();
}

inline bool
kernel_lock_t::test() const
{
  unsigned long flags = get_eflags();
  return !(flags & EFL_IF);
}

#else  // PROFILE

// When profiling, we use a sligtly different strategy: Instead of
// disabling interrupts in the CPU, we disable all interrupts but the
// profiling timer interrupt in the PIC.

inline
kernel_lock_t::kernel_lock_t()
  : _is_set(0)
{}

inline void
kernel_lock_t::lock()
{
  cli();
  if (! _is_set)
    {
      // mask out all irqs except the profiling timer interrupt
      unsigned profile_irq_mask = 1 << config::profile_irq;

      _mask_master = inb(MASTER_OCW);
      _mask_slave =  inb(SLAVES_OCW);
      outb(MASTER_OCW, 0xff ^ (profile_irq_mask & ~_mask_master));
      outb(SLAVES_OCW, 0xff ^ ((profile_irq_mask >> 8) & ~_mask_slave));

      _is_set = true;
    }
  sti();
  asm("nop; nop");		// last chance for an irq to be delivered
}

inline void
kernel_lock_t::clear_irqdisable()
{
  cli();
  if (_is_set)
    {
      outb(MASTER_OCW, _mask_master);
      outb(SLAVES_OCW, _mask_slave);

      _is_set = false;
    }
}

inline void
kernel_lock_t::clear()
{
  clear_irqdisable();
  sti();
}

inline bool
kernel_lock_t::test() const
{
  return _is_set;
}

#endif // PROFILE

inline bool
kernel_lock_t::test_and_set()
{
  bool ret = test();
  lock();
  return ret;
}

inline void
kernel_lock_t::set(bool state)
{
  if (state)
    lock();
  else
    clear();
}

#endif // KERNEL_LOCK_H
