// kmem_slab_t -- A type-independent slab cache allocator for Fiasco,
// derived from a generic slab cache allocator (slab_cache_anon in
// lib/slab.{h,cc}).

// This specialization adds low-level page allocation and locking to
// the slab allocator implemented in our base class (slab_cache_anon).
// We allocate multi-page slabs as multiple, potentially
// non-contiguous physical pages and map them contiguously into a
// region of our virtual address space managed by region_t (in
// region.{h,cc}).

// XXX This implementation technique has one important drawback: When
// using virtual pages, we need TLB entries for these 4KB pages, thus
// increasing the number of 4KB-page TLB misses, whereas accesses to
// physical pages would use superpage (4MB) mappings using the
// 4MB-page TLB.  The load on the superpage TLB is substantially
// lower.  Maybe we should try to find a contiguous physical region
// first and only use virtual regions as a last resort?

#include <assert.h>

#include "kmem.h"
#include "region.h"
#include "lock.h"

#include "kmem_slab.h"		// the interface we implement

kmem_slab_t::kmem_slab_t(unsigned long slab_size, unsigned elem_size, 
	    unsigned alignment)
  : slab_cache_anon(slab_size, elem_size, alignment)
{
  static bool region_initialized = false;

  if (!region_initialized)
    {
      region_initialized = true; // set this first to avoid recursion
      region::init();		 // because region allocs a slab of its own
    }
}

kmem_slab_t::~kmem_slab_t()
{
  helping_lock_guard_t guard(&_lock);
  destroy();
}

// We overwrite some of slab_cache_anon's functions to faciliate locking.
void *
kmem_slab_t::alloc()
{
  helping_lock_guard_t guard(&_lock);
  return slab_cache_anon::alloc();
}

void 
kmem_slab_t::free(void *cache_entry)
{
  helping_lock_guard_t guard(&_lock);
  slab_cache_anon::free(cache_entry);
}

bool 
kmem_slab_t::reap()
{
  if (_lock.test_and_set())
    return false;		// this cache is locked -- can't get memory now

  bool ret = slab_cache_anon::reap();
  _lock.clear();
  return ret;
}

// Callback functions called by our super class, slab_cache_anon, to
// allocate or free blocks

void *
kmem_slab_t::block_alloc(unsigned long size, unsigned long alignment)
{
  // size must be a power of two of PAGE_SIZE
  assert(size >= PAGE_SIZE
	 && (size & (size - 1)) == 0); // only one bit set -> power of two

  // If the size is just one page, just allocate a page from kmem; do
  // not reserve a region for it.  This is not just an optimization:
  // As the region_t implementation internally uses a kmem_slab_t
  // object itself, it relies on this property.

  if (size == PAGE_SIZE
      && alignment <= PAGE_SIZE)
    {
      return kmem::page_alloc();
    }

  vm_offset_t vaddr = region::reserve_pages(size, alignment);
  if (! vaddr)
    return 0;

  // Fine, we reserved virtual addresses for the buffer.  Now actually
  // allocate pages.
  for (vm_offset_t a = vaddr;
       a < vaddr + size;
       a += PAGE_SIZE)
    {
      if (kmem::page_alloc(a, kmem::zero_fill))
	continue;		// successfully allocated a page

      // Error - out of memory.  Undo everything.
      for (vm_offset_t i = vaddr;
	   i < a;
	   i += PAGE_SIZE)
	{
	  kmem::page_free(reinterpret_cast<void*>(i));
	}

      region::return_pages(vaddr, size);
      
      return 0;			// return error
    }

  return reinterpret_cast<void*>(vaddr);
}

void 
kmem_slab_t::block_free(void *block, unsigned long size)
{
  for (char *p = reinterpret_cast<char*>(block);
       p < reinterpret_cast<char*>(block) + size;
       p += PAGE_SIZE)
    {
      kmem::page_free(p);
    }

  // We didn't reserve a memory region if the allocation was just one
  // page.  Otherwise, we need to free the region.
  if (size != PAGE_SIZE)	
    region::return_pages(reinterpret_cast<vm_offset_t>(block), size);
}

// memory management

void *
kmem_slab_t::operator new(size_t size)
{
  return kmem::alloc(size);
}

void 
kmem_slab_t::operator delete(void *block, size_t size)
{
  kmem::free(block, size);
}
