#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <malloc.h>

#include <flux/gdb_serial.h>
#include <flux/tty.h>
#include <flux/machine/pc/direct_cons.h>
#include <flux/machine/pc/reset.h>
#include <flux/machine/pc/com_cons.h>

#include "lib.h"
#include "config.h"
#include "kdb.h"
#include "jdb.h"

static void serial_init();
static bool serial_initialized = false;


extern "C" void
serial_putchar(int c)
{
  if (config::use_serial)
    {
      if (kdb::connected())
	{
	  if (jdb::source_level_debugging())
	    gdb_serial_putchar(c);
	}
      else
	{
	  if (! serial_initialized)
	    serial_init();

	  com_cons_putchar(c);
	  if (c == '\n')
	    com_cons_putchar('\r');
	}
    }
}


int direct_cons_try_getchar(void);
//int com_cons_try_getchar();

extern "C" int
getchar(void)
{

  // XXX should poll both serial and local console at the same time
  // XXX can't get characters from a GDB console

  int c = -1;

  while (c == -1) {
    if (config::use_serial && ! kdb::connected())
      {
	if (! serial_initialized)
	  serial_init();
	
	c = com_cons_try_getchar();
      }
    if (c == -1)
      c = direct_cons_try_getchar();
  }
  return c;
}

extern "C" void 
_exit(int)
{
  char c;

  cli();
  printf("\nReturn reboots, \"k\" enters L4 kernel debugger...\n");

  c = getchar();

  if (c == 'k' || c == 'K') 
    kdb::ke("_exit");

  pc_reset();
}

static void 
serial_init()
{
  com_cons_init(config::serial_com_port, &base_raw_termios);
  serial_initialized = true;
  printf("serial_init\n");
}


// from oskit/libkern/x86/pc/direct_cons_putchar.c -- adapted for Hercules

/* 
 * Copyright (c) 1995-1994 The University of Utah and
 * the Computer Systems Laboratory at the University of Utah (CSL).
 * All rights reserved.
 *
 * Permission to use, copy, modify and distribute this software is hereby
 * granted provided that (1) source code retains these copyright, permission,
 * and disclaimer notices, and (2) redistributions including binaries
 * reproduce the notices in supporting documentation, and (3) all advertising
 * materials mentioning features or use of this software display the following
 * acknowledgement: ``This product includes software developed by the
 * Computer Systems Laboratory at the University of Utah.''
 *
 * THE UNIVERSITY OF UTAH AND CSL ALLOW FREE USE OF THIS SOFTWARE IN ITS "AS
 * IS" CONDITION.  THE UNIVERSITY OF UTAH AND CSL DISCLAIM ANY LIABILITY OF
 * ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * CSL requests users of this software to return to csl-dist@cs.utah.edu any
 * improvements that they make and grant CSL redistribution rights.
 */

/* This is a special "feature" (read: kludge)
   intended for use only for kernel debugging.
   It enables an extremely simple console output mechanism
   that sends text straight to CGA/EGA/VGA video memory.
   It has the nice property of being functional right from the start,
   so it can be used to debug things that happen very early
   before any devices are initialized.  */

#include <flux/c/string.h>
#include <flux/base_critical.h>
#include <flux/x86/base_vm.h>
#include <flux/x86/pc/direct_cons.h>




#define HERC_VIDBASE	((unsigned char *)phystokv(0xb0000))
#define VGA_VIDBASE	((unsigned char *)phystokv(0xb8000))

static unsigned char * vid_base=VGA_VIDBASE;

extern "C" void
switch_vid(int herc)
{
//   printf("switching to %d\n", herc);
  if (herc)
    vid_base=HERC_VIDBASE;
  else
    vid_base=VGA_VIDBASE;
}


void
vt100(unsigned char c)
{
  serial_putchar(27);
  serial_putchar('[');
  serial_putchar(c);
}

extern "C" int
putchar(int c)
{
  static int cursor_x = -1;
  static int cursor_y = 24;
  static int charmode = 0;
  static char tmp_buff[15];

  base_critical_enter();

  if (cursor_x < 0)
    {
      /* Called for the first time - initialize.  */
      cursor_x = 0;
      putchar('\n');
    }

  if (charmode == 1) {
    cursor_y = c;
    charmode = 2;
  }
  else if (charmode == 2) {
    char *p;
	  
    cursor_x = c;
    charmode=0;
    /* execute vt100 command ^[x;yH */
    sprintf(tmp_buff, "%c[%d;%dH", (char)27, cursor_y+1, cursor_x+1);
    for(p=tmp_buff; *p; p++)
      serial_putchar(*p);
  }
  else {
    switch (c)
      {
      case '\n':
	if (cursor_y == 24) {
	  bcopy(vid_base+80*2, vid_base, 80*2*24);
	  bzero(vid_base+80*2*24, 80*2);
	}
	else
	  cursor_y++;
	cursor_x = 0;
	serial_putchar('\n');
	serial_putchar('\r');
	break;

      case '\r':
	cursor_x = 0;
	serial_putchar('\r');
	break;
	      
      case '\t':
	cursor_x = (cursor_x + 8) & ~7;
	serial_putchar('\t');
	break;
	      
      case 6:
	/* cursor */
	charmode = 1;
	break;

      case 1:
	/* home */
	cursor_x = cursor_y = 0;
	vt100('H');
	break;

      case 5:
	/* clear to end of line */
	{
	  unsigned char *p = vid_base + 80*2*cursor_y;
	  int i;
	  for (i=cursor_x*2; i < 80 *2; i++) {
	    p[i] = '\0';
	    p[i+1] = 7;
	  }
	  vt100('K');
	  break;
	}

      case 8:
	/* back space */
	if (cursor_x)
	  cursor_x--;
	serial_putchar(8);
	break;

      default:
	/* Wrap if we reach the end of a line.  */
	if (cursor_x >= 80) {
	  putchar('\n');
	}

	/* Stuff the character into the video buffer. */
	{
	  volatile unsigned char *p = vid_base + 80*2*cursor_y + cursor_x*2;
	  p[0] = c;
	  p[1] = 0x0f;
	  cursor_x++;
	  serial_putchar(c);
	}
	break;
      }
  }
  base_critical_leave();
  return c;
}


/* direct_cons_try_getchar() */

#include <flux/x86/pio.h>
#include <flux/x86/pc/keyboard.h>

#define SHIFT -1

static char keymap[128][2] = {
  {0},			/* 0 */
  {27,	27},		/* 1 - ESC */
  {'1',	'!'},		/* 2 */
  {'2',	'@'},
  {'3',	'#'},
  {'4',	'$'},
  {'5',	'%'},
  {'6',	'^'},
  {'7',	'&'},
  {'8',	'*'},
  {'9',	'('},
  {'0',	')'},
  {'-',	'_'},
  {'=',	'+'},
  {8,	8},		/* 14 - Backspace */
  {'\t',	'\t'},		/* 15 */
  {'q',	'Q'},
  {'w',	'W'},
  {'e',	'E'},
  {'r',	'R'},
  {'t',	'T'},
  {'y',	'Y'},
  {'u',	'U'},
  {'i',	'I'},
  {'o',	'O'},
  {'p',	'P'},
  {'[',	'{'},
//   {']',	'}'},		/* 27 */
  {'+',	'*'},		/* 27 */
  {'\r',	'\r'},		/* 28 - Enter */
  {0,	0},		/* 29 - Ctrl */
  {'a',	'A'},		/* 30 */
  {'s',	'S'},
  {'d',	'D'},
  {'f',	'F'},
  {'g',	'G'},
  {'h',	'H'},
  {'j',	'J'},
  {'k',	'K'},
  {'l',	'L'},
  {';',	':'},
  {'\'',	'"'},		/* 40 */
  {'`',	'~'},		/* 41 */
  {SHIFT,	SHIFT},		/* 42 - Left Shift */
  {'\\',	'|'},		/* 43 */
  {'z',	'Z'},		/* 44 */
  {'x',	'X'},
  {'c',	'C'},
  {'v',	'V'},
  {'b',	'B'},
  {'n',	'N'},
  {'m',	'M'},
  {',',	'<'},
  {'.',	'>'},
  //  {'/',	'?'},		/* 53 */
  {'-',	'_'},		/* 53 */
  {SHIFT,	SHIFT},		/* 54 - Right Shift */
  {0,	0},		/* 55 - Print Screen */
  {0,	0},		/* 56 - Alt */
  {' ',	' '},		/* 57 - Space bar */
  {0,	0},		/* 58 - Caps Lock */
  {0,	0},		/* 59 - F1 */
  {0,	0},		/* 60 - F2 */
  {0,	0},		/* 61 - F3 */
  {0,	0},		/* 62 - F4 */
  {0,	0},		/* 63 - F5 */
  {0,	0},		/* 64 - F6 */
  {0,	0},		/* 65 - F7 */
  {0,	0},		/* 66 - F8 */
  {0,	0},		/* 67 - F9 */
  {0,	0},		/* 68 - F10 */
  {0,	0},		/* 69 - Num Lock */
  {0,	0},		/* 70 - Scroll Lock */
  {'7',	'7'},		/* 71 - Numeric keypad 7 */
  {'8',	'8'},		/* 72 - Numeric keypad 8 */
  {'9',	'9'},		/* 73 - Numeric keypad 9 */
  {'-',	'-'},		/* 74 - Numeric keypad '-' */
  {'4',	'4'},		/* 75 - Numeric keypad 4 */
  {'5',	'5'},		/* 76 - Numeric keypad 5 */
  {'6',	'6'},		/* 77 - Numeric keypad 6 */
  {'+',	'+'},		/* 78 - Numeric keypad '+' */
  {'1',	'1'},		/* 79 - Numeric keypad 1 */
  {'2',	'2'},		/* 80 - Numeric keypad 2 */
  {'3',	'3'},		/* 81 - Numeric keypad 3 */
  {'0',	'0'},		/* 82 - Numeric keypad 0 */
  {'.',	'.'},		/* 83 - Numeric keypad '.' */
};

int
direct_cons_try_getchar(void)
{
  static unsigned shift_state;
  unsigned status, scan_code, ch;

  base_critical_enter();

retry:

  /* Wait until a scan code is ready and read it. */
  status = inb(0x64);
  if ((status & K_OBUF_FUL) == 0) {
    base_critical_leave();
    return -1;
  }
  scan_code = inb(0x60);

  /* Handle key releases - only release of SHIFT is important. */
  if (scan_code & 0x80) {
    scan_code &= 0x7f;
    if (keymap[scan_code][0] == SHIFT)
      shift_state = 0;
    goto retry;
  }

  /* Translate the character through the keymap. */
  ch = keymap[scan_code][shift_state];
  if (ch == (unsigned)SHIFT) {
    shift_state = 1;
    goto retry;
  } else if (ch == 0)
    goto retry;

  base_critical_leave();

  return ch;
}

// do not allow calls to these OSKit routines.  We have to provide
// these functions because libamm references them.

extern "C" void *smalloc(size_t size)
{
  panic("smalloc called");
}

extern "C" void sfree(void *buf, size_t size)
{
  panic("sfree called");
}
