#include "kdb.h"
#include "globals.h"
#include "kmem.h"
#include "kmem_slab.h"
#include "lock.h"
#include "oskit/amm.h"

#include "region.h"		// the interface we implement

extern "C" {
  extern char _mappings_1, _mappings_end_1;
}

static const vm_offset_t mem_alloc_region
  = reinterpret_cast<vm_offset_t>(&_mappings_1);
static const vm_offset_t mem_alloc_region_end
  = reinterpret_cast<vm_offset_t>(&_mappings_end_1);

// 
// helpers for the address map library
// 

// There is an interesting cross-dependency between this
// implementation of region_t and the kmem_slab_t implementation (in
// kmem_slab.cc): region_t uses a slab cache of type kmem_slab_t to
// manage entries of the address map library.  kmem_slab_t uses
// region_t to keep track page mappings for multi-page slabs.  To
// avoid recursions, we have to make sure we only use single-page
// slabs in our slab cache -- kmem_slab_t can handle those independent
// of region_t.
static kmem_slab_t *amm_entry_cache;

static amm_entry_t *amm_alloc_func(amm_t *amm, oskit_addr_t addr, 
				   oskit_size_t size, int flags)
{
  return reinterpret_cast<amm_entry_t *>(amm_entry_cache->alloc());
}

static void amm_free_func(amm_t *amm, amm_entry_t *entry)
{
  amm_entry_cache->free(entry);
}

// 
// region
// 

static amm_t region_amm;
static oskit_addr_t end_of_last_region;
static helping_lock_t region_lock;

void
region::init()
{
  // Make sure our slab cache only uses single-page slabs (slab_size =
  // PAGE_SIZE).  See note above declaration of amm_entry_cache for
  // more information.
  amm_entry_cache = new kmem_slab_t(PAGE_SIZE, sizeof(amm_entry_t), 4);

  amm_init_gen(&region_amm, AMM_FREE, 0, amm_alloc_func, amm_free_func, 0, 0);
  check ( amm_modify(&region_amm, 0, mem_alloc_region, AMM_RESERVED, 0) == 0 );
  check ( amm_modify(&region_amm, mem_alloc_region_end, 
		     AMM_MAXADDR - mem_alloc_region_end, 
		     AMM_RESERVED, 0) == 0 );

  end_of_last_region = mem_alloc_region;
}
 

vm_offset_t
region::reserve_pages(vm_size_t size, unsigned long alignment)
{
  helping_lock_guard_t guard(&region_lock);

  oskit_addr_t address = end_of_last_region;

  int align_bits = 0;
  while ((alignment >>= 1) != 0)
    align_bits++;

  if (! amm_find_gen(&region_amm, &address, size, AMM_FREE, -1, align_bits,
		     0, 0))
    {
      return 0;			// nothing found
    }

  end_of_last_region = address + size;

  if (amm_modify(&region_amm, address, size, AMM_ALLOCATED, 0) != 0)
    {
      return 0;			// error
    }

  return address;
}

void 
region::return_pages(vm_offset_t address, vm_size_t size)
{
  helping_lock_guard_t guard(&region_lock);

  assert( amm_find_gen(&region_amm, &address, size, AMM_ALLOCATED, -1,
		       0, 0, AMM_EXACT_ADDR) );

  check ( amm_modify(&region_amm, address, size, AMM_FREE, 0) != 0);
}

void
region::debug_dump()
{
  amm_dump(& region_amm);
}
