#include <l4/sys/types.h>
#include <flux/machine/pc/rtc.h>
#include <flux/machine/proc_reg.h>
#include <flux/machine/base_trap.h>

#include <string.h>
#include <stdio.h>

#include "config.h"
#include "globals.h"
#include "kdb.h"
#include "jdb.h"
#include "time.h"
#include "ipc.h"
#include "signal.h"
#include "profile.h"

#include "irq.h"
#include "thread.h"
#include "thread_list.h"

//
// class thread_t
//

thread_t *thread_t::prio_first[256];
thread_t *thread_t::prio_next[256];
unsigned thread_t::prio_highest;

int (*thread_t::nested_trap_handler)(trap_state *state) = 0;

int timeslice_ticks_left = config::default_time_slice; 

// allocator
// operator new() in thread.h

// deallocator
void 
thread_t::operator delete(void *)
{
  // XXX should check if all thread blocks on a given page are free
  // and deallocate (or mark free) the page if so.  this should be
  // easy to detect since normally all threads of a given task are
  // destroyed at once at task deletion time
}

// constructor

// the constructor doesn't really initialize the data members of of
// the thread control block (tcb) ; instead, it sets up the the tcb
// just enough so that the thread can be sent signals
thread_t::thread_t()
{
  // being able to receive signals involves: (a) push some restart
  // address on stack (we use a do-nothing routine which just always
  // schedules); (b) setting up the thread's the kernel stack pointer

  unsigned long *init_sp = reinterpret_cast<unsigned long*>
    (reinterpret_cast<unsigned long>(this) + size - sizeof(thread_regs_t));

  // set a magic value -- we use it later to verify the stack hasn't
  // been overrun
  _magic = magic;

  // don't care about errors: they just mean someone else has already
  // set up the tcb
  compare_and_swap(reinterpret_cast<void(**)()> (--init_sp), 
		   reinterpret_cast<void(*)()> (0), 
		   user_invoke);
  compare_and_swap(reinterpret_cast<unsigned long*>(&kernel_sp),
		   0UL, reinterpret_cast<unsigned long>(init_sp));

  // ok, we're ready to go!
}

thread_t::~thread_t()
{
  assert(state() == Thread_dead);

  assert(! my_signals.first());
  assert(_magic == magic);

  unsigned long *init_sp = reinterpret_cast<unsigned long*>
    (reinterpret_cast<unsigned long>(this) + size - sizeof(thread_regs_t));

  switching_in.clear();

  kernel_sp = 0;
  *--init_sp = 0;
  my_state = Thread_invalid;
}

// public services

// the scheduler
void 
thread_t::schedule()
{
  // Careful!  At this point, we may (due to a context switch) be
  // dequeued from the ready list but then set ready again, so
  // that our "ready_next" pointer becomes invalid
  kernel_lock.lock();
  my_ticks_left = timeslice_ticks_left ? timeslice_ticks_left : my_timeslice;

  if (in_ready_list())
    {
      if (! (my_state & Thread_running))
	ready_dequeue();
      else
	prio_next[my_prio] = ready_next->my_prio == my_prio ? 
	  ready_next : prio_first[my_prio];
    }
  else    
    {
      // we're not in the ready list.  if we're ready nontheless,
      // enqueue now so that we can be scheduled correctly

      if (my_state & Thread_running)
	ready_enqueue();
    }

  thread_t *next_to_run;

  for (;;)
    {
      next_to_run = prio_next[prio_highest];

      assert (next_to_run);

      if (next_to_run->my_state & Thread_running) 
	break;

      next_to_run->ready_dequeue();

      kernel_lock.clear();
      asm("nop; nop");		// give CPU a chance to accept interrupts
      kernel_lock.lock();
    }
	
  timeslice_ticks_left = next_to_run->my_ticks_left;

  if (next_to_run != this) switch_to(next_to_run);
  kernel_lock.clear();
}

//
// private functions 
//

bool 
thread_t::do_signals()
{
  // make sure state-switch handling and all signals are serialized
  if (switching_in.test_and_set())
    return false;

  for (;;)
    {
      // Not a nested invocation -- it's safe to unlock kernel now.
      // (See comment on switchin_context().)
      bool was_locked = kernel_lock.test();
      kernel_lock.clear();

      // check for signals
      for (signal_t *s = my_signals.dequeue(); 
	   s; 
	   s = my_signals.dequeue())
	{
	  thread_t *sender = s->sender();
	  assert(_current_trans == 0);
	  _current_trans = s;
	  unsigned ret = s->func(this); // call signal handler

	  // Order is important as we might be killed immediately
	  // after calling set_ret().
	  _current_trans = 0;
	  s->set_ret(ret);

	  unsigned hint = s->switch_hint();
	  
	  s = my_signals.first(); // next signal?
	  
	  if (s 		// there is another signal and it
	      && sender->my_prio < s->sender()->my_prio) // has higher prio
	    continue;		// then handle next signal immediately
	    
	  // switch back so sender of signal if we're not ready to run
	  // and higher priorized
	  if (this != sender)	// it's not ourself -- could switch back
	    {
	      if ((! (state() & Thread_running)) // must switch, cannot run!
		  || (hint != signal_t::Ret_activate_receiver
		      && (s	// more signals left -- not ready to run
			  || sender->my_prio >= my_prio))
		  || hint == signal_t::Ret_activate_sender)
		{
		  // then switch to sender of handled signal
		  switch_to(sender);
		}
	    }
	}

      while ((! my_signals.first())
	     && (! (state() & Thread_running)))
	{
	  schedule();
	}

      // if we were in uninterruptible mode, become uninterruptible again
      kernel_lock.set(was_locked);

      switching_in.clear();

      // repeat if more signals arrived

      if (! my_signals.first())
	break;			
  
      check (switching_in.test_and_set() == 0);
    }

  return true;
}


// send a signal to a thread
unsigned long 
thread_t::send_signal(thread_t *to, signal_t *s)
{
  assert(this == current());
  assert((state() & Thread_running) || switching_in.test());

  // if this is the first signal we're sending (i.e., we're not
  // sending a signal from a signal handler), register the signal

  bool first_signal = state_add(Thread_signalling);

  if (first_signal)
    {
      // save state so that we can be deleted safely
      my_signal_partner = to; // save the partner we send signal to
      my_sent_signal = s;
    }

  // insert signal structure into target's signal list
  if (! to->insert_signal(s))
    {
      if (first_signal)
	{
	  my_sent_signal = 0;
	  state_del(Thread_signalling);
	}
      s->set_ret(signal_t::Sig_aborted);
      return signal_t::Sig_aborted;
    }

  // execute the signal handler in target's context
  while (! s->done()) 
    {
      if (to == this)
	{
	  // do signals locally
	  if (! switchin_context())
	    panic("recursive signals");	// we interrupted a signal handler
	}
      else
	{
	  if (! switch_to(to))
	    {
	      s->set_ret(signal_t::Sig_aborted);
	      break;
	    }
	}
    }

  // done -- resume execution as normal
  if (first_signal)
    {
      my_sent_signal = 0;
      state_del(Thread_signalling);
    }
  return s->ret();
}

// Thread context switchin -- called on every re-activation of a
// thread (switch_to()) and whenever we had acquired the switching_in
// lock.  When called from switch_to(), we have the kernel lock and
// must unlock the kernel when appropriate (that is, only if this is
// not a nested call -- that is, if we could lock the switching_in
// lock).
bool
thread_t::switchin_context()
{
  assert(this == current());

  // do_signals() handles the sti() magic explained above

  if (! do_signals())		// could we deliver signals (sig q lockable?)
    {				// no -- signal processing already in progress
      assert(switching_in.test());

      return false;
    }

  // only need to do more switching work if we were able to lock the
  // signal queue

#if 0  
  assert(state() & Thread_running); // XXX actually, this can happen;
				    // we should schedule...
#endif

  // set kernel-esp in case we want to return to the user
  *(kmem::kernel_esp()) = reinterpret_cast<vm_offset_t>(regs() + 1);
  
  // switch to our page directory if nessecary
  my_space->switchin_context();

  return true;
}

// switch to a specific different thread and immediately handle its signals
bool 
thread_t::switch_to(thread_t *t)
{
  assert(current() != t);
  assert(this == current());

  bool was_locked = kernel_lock.test_and_set();
  // XXX we assume here that irqs are at least disabled on this CPU --
  // valid for all implementations of kernel_lock?

  // Make sure stack has enough space.  (Irqs must be disabled here
  // because otherwise we could be preempted between the tests, and
  // t->kernel_sp could be set to 0 after we checked for that case.)
  assert(t->kernel_sp == 0 
	 || reinterpret_cast<vm_offset_t>(t->kernel_sp)
	    > reinterpret_cast<vm_offset_t>(t) + sizeof(thread_t) + 0x20);

  // Save FPU state now -- don't do this lazily (see comment on
  // handle_fpu_trap() below)
  if (state() & Thread_fpu_active)
    {
      asm volatile("fnsave %0 ; fwait":"=m" (fpu_regs));
      state_del(Thread_fpu_active);
      set_ts();
    }

  if ((state() & Thread_running) && ! in_ready_list())
    ready_enqueue();

  // switch to new thread's stack

  bool ret;

  // caution: be sure not to use any registers in here so that the
  // thread switching code remains preemptible as well
  asm volatile
    (// save context of old thread
#ifndef NO_FRAME_PTR
     "   pushl %%ebp \n"
#endif
     "   pushl $1f \n"		// push restart address on old stack

     "   movl  %%esp, (%1) \n"	// save stack pointer

     // read context of new thread
     "   movl  (%2), %%esp \n"	// load new stack pointer - now in other thread
     "   testl %%esp,%%esp \n"	// check new stack pointer - is it 0?
     "   jz    3f \n"		// yes - fail

     // in new context now (still cli'd)

     // deliver signals to new thread
     "   pushl %%esp \n"	// pass ptr to context as argument
     "   pushl %3 \n"		// new thread's "this" -- XXX correct?
     "   call  %4 \n"		// call switchin_context()
     "   addl  $8,%%esp \n"	// clean up args

     // return to old context
     "   ret \n"		// jump to restart address on stack

     // code for failure
     "3: \n"
     "   movl  (%1), %%esp \n"	// load old stack pointer
     "   addl  $4, %%esp \n"
     "   xor   %0,%0 \n"	// return false
     "   jmp   2f \n"

     // label for our restart address
     "   .p2align 4 \n"		// start code at new cache line
     "1: mov   $1,%0 \n"	// return true
     "2: \n"			// label for end of this asm segment
#ifndef NO_FRAME_PTR
     "   popl %%ebp \n"
#endif
     : "=a" (ret)
     : "r" (&kernel_sp), "r" (&t->kernel_sp), "r" (t), "m" (switchin_context)
     : "ebx", "ecx", "edx", "esi", "edi", 
#ifdef NO_FRAME_PTR
       "ebp", 
#endif
       "memory");

  kernel_lock.set(was_locked);

  return ret;
}

//
// internal stuff -- all of these routines are serialized and run on
// the stack of the thread to be manipulated, so we don't need to
// synchronize access to private members (except for the kernel stack
// pointer)
//

// routines which are implemented as signals should be inline
unsigned 
thread_t::reset(space_t* space,
		const l4_threadid_t* id, 
		int init_prio, unsigned short mcp)
{
  if (! in_transaction())	// not already in signal?
    {
      return trans_invoke(this, &reset, space, id, init_prio, mcp);
    }

  if (state() != Thread_invalid)
    return signal_t::Sig_fail;

  // make sure we're freshly initialized
  assert(*reinterpret_cast<void(**)()>(reinterpret_cast<unsigned long>(this) 
				       + thread_t::size 
				       - sizeof(thread_regs_t) - 4)
	 == thread_t::user_invoke);

  my_space = space;
  my_id = *id;
  my_prio = init_prio;
  my_mcp = mcp;
  my_irq = 0;
  my_idt_limit = 0;

  my_timeslice = config::default_time_slice;
  my_ticks_left = config::default_time_slice;
  
  // clear out user regs that can be returned from the thread_ex_regs
  // system call to prevent covert channel
  thread_ret_regs_t *r = regs();
  r->esp = 0;
  r->eip = 0;
  r->eflags = 0;

  // make sure the thread's kernel stack is mapped in its address space
  my_space->kmem_update(reinterpret_cast<vm_offset_t>(this));
  
  state_add(Thread_dead);
  
  my_pager = my_preempter = my_ext_preempter = nil_thread;

  if (space_index() == _current_trans->sender()->space_index())
    {
      // same task -> enqueue after creator
      present_enqueue(_current_trans->sender());
    }
  else
    {
      // other task -> enqueue in front of this task
      present_enqueue(_current_trans->sender()->
		      space_index().lookup_thread()->present_prev);
      // that's safe because thread 0 of a task is always present
    }

  return signal_t::Sig_success;
}

unsigned
thread_t::initialize(vm_offset_t eip, vm_offset_t esp,
		     thread_t* pager, thread_t* preempter,
		     vm_offset_t *o_eip = 0, 
		     vm_offset_t *o_esp = 0,
		     thread_t* *o_pager = 0, 
		     thread_t* *o_preempter = 0,
		     vm_offset_t *o_eflags = 0)
{
  if (! in_transaction())	// not already in signal?
    {
      return trans_invoke(this, &initialize, eip, esp, pager, preempter,
			  o_eip, o_esp, o_pager, o_preempter, o_eflags);
    }
  
  if (state() == Thread_invalid)
    return signal_t::Sig_fail;

  thread_ret_regs_t *r = regs();

  if (o_pager) *o_pager = my_pager;
  if (o_preempter) *o_preempter = my_preempter;
  if (o_esp) *o_esp = r->esp;
  if (o_eip) *o_eip = r->eip;
  if (o_eflags) *o_eflags = r->eflags;

  if (eip != 0xffffffff)
    {
      r->eip = eip;
      if (! (state() & Thread_dead))
	{
#if 0
	  kdb::ke("reseting non-dead thread");
#endif
	  // cancel ongoing IPC or other activity
	  state_change(~Thread_ipc_in_progress, Thread_cancel);
	}
    }

  if (pager != 0) my_pager = pager;
  if (preempter != 0) my_preempter = preempter;
  r->cs = kmem::gdt_code_user | SEL_PL_U;
  r->eflags = EFL_IOPL_USER | EFL_IF | 2;	// XXX iopl=kernel, ei
  if (esp != 0xffffffff) r->esp = esp;
  r->ss = kmem::gdt_data_user | SEL_PL_U;
  
  state_change(~Thread_dead, Thread_running);
  // need not enqueue in ready queue here -- switch_to() will do this for us
  
  return signal_t::Sig_success;
}

unsigned 
thread_t::set_sched_param(l4_sched_param_t param, thread_t* ext_preempter, 
			  l4_sched_param_t *o_param, l4_low_high_t *time,
			  thread_t **o_ext_preempter, sender_t **partner)
{
  if (! in_transaction())	// not already in signal?
    {
      return trans_invoke(this, &set_sched_param, param, ext_preempter, 
			  o_param, time, o_ext_preempter, partner);
    }
  
  if (state() == Thread_invalid)
    return signal_t::Sig_fail;

  // call legal?
  if (my_prio > _current_trans->sender()->my_mcp)
    return signal_t::Sig_fail;

  // query current state
  unsigned s = state();

  if (s & (Thread_polling|Thread_receiving))
    *partner = my_partner;
  else
    *partner = 0;

  if (s & Thread_dead) s = 0xf;
  else if (s & Thread_polling) s = (s & Thread_running) ? 8 : 0xd;
  else if (s & (Thread_waiting|Thread_receiving)) 
    s = (s & Thread_running) ? 8 : 0xc;
  else s = 0;

  o_param->sp.zero = s;

  unsigned long long timeslice = my_timeslice * config::microsec_per_tick;
  unsigned exp = 15;

  while (timeslice > 255)
    {
      timeslice >>= 2;
      exp--;
    }

  o_param->sp.time_exp = exp;
  o_param->sp.time_man = timeslice;

  o_param->sp.prio = my_prio;

  o_param->sp.small = 0;	// unsupported

  *o_ext_preempter = my_ext_preempter;

  time->low = time->high = 0;	// XXX unsupported

  // set new state
  if (param.sched_param != 0xffffffff)
    {
      if (param.sp.prio <= _current_trans->sender()->my_mcp)
	if (my_prio != param.sp.prio)
	  {
	    // We need to protect the priority manipulation so that
	    // this thread can not be preempted an ready-enqueued
	    // according to a wrong priority
	    kernel_lock.lock();
	    if (in_ready_list())
	      ready_dequeue();	// need to re-queue in ready queue
				// according to new prio
	    my_prio = param.sp.prio;
	    kernel_lock.clear();
	  }
      
      if (param.sp.time_exp)
	{
	  if (param.sp.time_man)
	    {
	      timeslice = static_cast<unsigned long long>(param.sp.time_man)
		<< ((15 - param.sp.time_exp) << 1);
	      my_timeslice = timeslice / config::microsec_per_tick;

	      if (my_timeslice == 0)
		my_timeslice = 1;
	    }
	  else
	    my_timeslice = 0;
	}
    }

  if (ext_preempter)
    {
      if (ext_preempter->exists())
	my_ext_preempter = ext_preempter;
      else
	kdb::ke("sig_sched: invalid ext_preempter");
    }

  return signal_t::Sig_success;
}

unsigned
thread_t::raise_exception(trap_state *ts, vm_offset_t handler)
{
  if (! in_transaction())	// not already in signal?
    {
      return trans_invoke(this, &raise_exception, ts, handler);
    }
  
  if (! (state() & Thread_running)
      || (state() & Thread_cancel))
    return signal_t::Sig_fail;

  ts->eip = handler;
  ts->esp -= ((ts->trapno >= 0x0a && ts->trapno <= 0x0e)
	      || ts->trapno == 0x08
	      || ts->trapno == 0x11) // need to push error code?
    ? 4 * 4
    : 3 * 4;

  return signal_t::Sig_success;
}

unsigned
thread_t::kill()
{
  if (! in_transaction())	// not already in signal?
    {
      return trans_invoke(this, &kill);
    }
  
  if (state() == Thread_invalid
      || (state() & Thread_dead))
    return signal_t::Sig_fail;

  assert(in_present_list());
  
  // If sending another signal, finish it before killing this thread
  if ((state() & (Thread_signalling|Thread_running))
      && my_sent_signal)
    {
      assert(my_signal_partner != this);

      while (! my_sent_signal->done())
	{
	  if (! switch_to(my_signal_partner))
	    break;		// could not switch into sig rcvr's context
	}
    }

  // If we're thread 0, start by deleting all our threads.

  if (id().id.lthread == 0)
    {
      // Kill threads belonging to the same task we find in the
      // present list.  We assume here that new threads of the same
      // task are always enqueued /after/ thread 0.
      while (present_next != this
	     && present_next->space() == space())
	{
	  // save in var because present_next will be dequeued from
	  // present queue
	  thread_t *n = present_next;
	  if (n->kill() != signal_t::Sig_success)
	    return signal_t::Sig_fail;
	  
	  delete n;
	}
      
      // Now that our brothers can't create new tasks any more, kill
      // subtasks.  We assume here that subtasks of a task are always
      // enqueued /before/ thread 0 of a task.
      
      while (present_prev != this
	     && present_prev->chief_index() == space_index())
	{
	  // save in var because present_prev will be dequeued from
	  // present queue
	  thread_t *n = present_prev->space_index().lookup_thread();
				// this yields thread 0 of that task
	  if (n->kill() != signal_t::Sig_success)
	    return signal_t::Sig_fail;
	  
	  delete n;
	}      

      //
      // Flush our address space
      //

      // It might be faster to flush our address space before actually
      // deleting the subtasks: Under the assumption that child tasks
      // got many of their mappings from this task, flushing mappings
      // early will flush their mappings as well and speed up their
      // deletion.  However, this would yield increased uglyness
      // because subtasks could still be scheduled and do page faults.
      // XXX check /how/ ugly...

      fpage_unmap(space(), l4_fpage(0, L4_WHOLE_ADDRESS_SPACE, 0, 0),
		  true, false);
    }

  //
  // Kill this thread
  //

  state_change(0, Thread_dead);

  // if attached to irqs, detach
  if (my_irq)
    my_irq->free(this);

  // if timeout active, abort
  if (my_timeout)
    my_timeout->reset();

  // if other threads are sending me signals, abort those
  for (signal_t *s = my_signals.dequeue(); 
       s; 
       s = my_signals.dequeue())
    {
      s->set_ret(signal_t::Sig_aborted);
    }

  // XXX what happens with signals that are enqueued from now on?

  // if other threads want to send me IPC messages, abort these
  // operations 

  // XXX that's quite difficult: how to traverse a list that may
  // change at all times?

  // if engaged in IPC operation, stop it
  if (in_sender_list())
    sender_dequeue(my_send_partner);
  
  // dequeue from system queues
  assert(in_present_list());
  present_dequeue();

  // XXX (don't need to dequeue from ready list -- this will happen
  // automagically when we're descheduled)
  if (in_ready_list())
    ready_dequeue();

  return signal_t::Sig_success;
}

unsigned 
thread_t::kill_task(space_index_t subtask)
{
  if (! in_transaction())	// not already in signal?
    {
      return trans_invoke(this, &kill_task, subtask);
    }
  
  if (state() == Thread_invalid
      || (state() & Thread_dead))
    return signal_t::Sig_fail;
  
  assert(id().id.lthread == 0);

  assert(subtask.chief() == space_index());

  thread_t *c = subtask.lookup_thread();

  if (c->kill() != signal_t::Sig_success)
    return signal_t::Sig_fail;

  delete c;

  return signal_t::Sig_success;
}



// helpers

void 
thread_t::user_invoke()
{
  while (! (current()->state() & Thread_running))
    current()->schedule();

  current()->state_del(Thread_cancel);

  asm volatile
    ("  movl %%eax,%%esp \n"	// set stack pointer to regs structure
     "  movw %1,%%ax \n"
     "  movw %%ax,%%es \n"
     "  movw %%ax,%%ds \n"
     "  xorl %%ecx,%%ecx\n"	// clean out user regs
     "  xorl %%edx,%%edx\n"
     "  xorl %%esi,%%esi\n"
     "  xorl %%edi,%%edi\n"
     "  xorl %%ebx,%%ebx\n"
     "  xorl %%ebp,%%ebp\n"
     "  xorl %%eax,%%eax\n"
     "  iret"
     :				// no output
     : "a" (current()->regs()), 
       "i" (kmem::gdt_data_user | SEL_PL_U)
     );

  // never returns here
}

// the global trap handler switch
bool
thread_t::handle_slow_trap(trap_state *ts)
{
  extern unsigned gdb_trap_recover; // in OSKit's gdb_trap.c
  unsigned eip;

  if (gdb_trap_recover)
    goto generic; // we're in the GDB stub -- let generic handler handle it

  if (! ((ts->cs & 3) || (ts->eflags & EFL_VM)))
    goto generic;		// we were in kernel mode -- nothing to emulate

  if (ts->trapno == 2)		// NMI?
    goto generic;		// NMI always enters kernel debugger

  if (ts->trapno == 0xffffffff)	// debugger interrupt
    goto generic;

  // so we were in user mode -- look for something to emulate

  sti();

  eip = ts->eip;

  // check for "lidt (%eax)"
  if (ts->trapno == 13 && (ts->err & 0xffff) == 0
      && eip < kmem::mem_user_max - 4
      && (*reinterpret_cast<unsigned long*>(eip) & 0xffffff) == 0x18010f)
    {
      // emulate "lidt (%eax)"

      // read descriptor
      if (ts->eax >= kmem::mem_user_max - 6)
	goto fail;

      x86_gate *idt = *reinterpret_cast<x86_gate**>(ts->eax + 2);
      vm_size_t limit = config::backward_compatibility ? 255
	: *reinterpret_cast<unsigned short*>(ts->eax);

      if (limit >= kmem::mem_user_max
	  || reinterpret_cast<vm_offset_t>(idt) >= kmem::mem_user_max - limit - 1)
	goto fail;

      // OK; store descriptor
      my_idt = idt;
      my_idt_limit = (limit + 1) / sizeof(x86_gate);

      // consume instruction and continue
      compare_and_swap(&ts->eip, eip, eip + 3); // ignore errors
      return true;
    }

  // let's see if we have a trampoline to invoke
  if (ts->trapno < 0x20
      && ts->trapno < my_idt_limit)
    {
      // translate page fault exceptions to general protection
      // exception -- users shouldn't have to care about about page
      // faults (and probably couldn't find out the fault address
      // anyway as they can't access the %cr3 (page fault address)
      // register)
      if (ts->trapno == 14)
	{
	  ts->trapno = 13;
	  ts->err = 0;
	}

      x86_gate *g = my_idt + ts->trapno;

      if (config::backward_compatibility // backward compat.: don't check
	  || ((g->word_count & 0xe0) == 0
	      && (g->access & 0x1f) == 0x0f)) // gate descriptor ok?
	{
	  vm_offset_t o = (g->offset_high << 16) | g->offset_low;

	  if ((o != 0 || !config::backward_compatibility) // bw compat.: != 0?
	      && o < kmem::mem_user_max // in user space?
	      && ts->esp <= kmem::mem_user_max
	      && ts->esp > 4 * 4) // enough space on user stack?
	    {
	      // OK, reflect the trap to user mode
	      unsigned32_t *esp = reinterpret_cast<unsigned32_t*>(ts->esp);
	      
	      if (raise_exception(ts, o) != signal_t::Sig_success)
		{
		  // someone interfered and changed our state 
		  check(state_del(Thread_cancel));

		  return true;
		}
	      
	      *--esp = ts->eflags;
	      *--esp = kmem::gdt_code_user | SEL_PL_U;
	      *--esp = eip;

	      if ((ts->trapno >= 0x0a
		   && ts->trapno <= 0x0e)
		  || ts->trapno == 0x08
		  || ts->trapno == 0x11) // need to push error code?
		{
		  *--esp = ts->err;
		}

	      return true;	// we've consumed the trap
	    }
	}
    }
  
  // backward compatibility cruft: check for those insane "int3" debug
  // messaging command sequences
  if (ts->trapno == 3 
      && (ts->eflags & EFL_IOPL) == EFL_IOPL_USER) // only allow priv tasks
    {
      // no bounds checking here -- we assume privileged tasks are
      // civilized citizens :-)
      unsigned char todo = *reinterpret_cast<char*>(eip);
      bool enter_kdb = false;

      char *str;
      int len;

      switch (todo)
	{
	case 0xeb:		// jmp == enter_kdebug()
	  printf("KDB: ");

	  len = *reinterpret_cast<char*>(eip + 1);
	  str = reinterpret_cast<char*>(eip + 2);

	  enter_kdb = true;
	  goto printstr;

	case 0x90:		// nop == kd_display()
	  if (*reinterpret_cast<unsigned char*>(eip + 1) != 0xeb /*jmp*/)
	    goto nostr;

	  len = *reinterpret_cast<char*>(eip + 2);
	  str = reinterpret_cast<char*>(eip + 3);

	printstr:
	  if (len <= 0)
	    goto nostr;

	  while (len--)
	    putchar(*str++);
      
	  break;

	case 0x3c:		// cmpb
	  todo = *reinterpret_cast<char*>(eip + 1);

	  switch (todo)
	    {
	    case 0:		// outchar
	      putchar(ts->eax & 0xff);
	      break;

	    case 2:		// outstring
	      printf("%s", reinterpret_cast<const char*>(ts->eax));
	      break;

	    case 5:		// outhex32
	      printf("%08x", ts->eax);
	      break;

	    case 6:		// outhex20
	      printf("%05x", ts->eax & 0xfffff);
	      break;

	    case 7:		// outhex16
	      printf("%04x", ts->eax & 0xffff);
	      break;

	    case 11:		// outdec
	      printf("%d", ts->eax);
	      break;

	    case 24:		// start kernel profiling
	      if (config::profiling)
		{
		  cli();
		  profile::start();
		  sti();
		}
	      break;

	    case 25:		// stop kernel profiling, dump data to serial
	      if (config::profiling)
		{
		  cli();
		  profile::stop_and_dump();
		  sti();
		}
	      break;

	    case 26:		// stop kernel profiling; do not dump
	      if (config::profiling)
		{
		  cli();
		  profile::stop();
		  sti();
		}
	      break;

	    default:		// ko
	      if (todo < ' ')
		goto nostr;

	      putchar(todo);
	    }	  

	  break;

	default:		// user breakpoint
	  goto nostr;
	}
      
      if (enter_kdb)
	{
	  if (kdb::connected() || jdb::connected())
	    {
	      printf("\n");
	      goto generic;
	    }

	  // no kernel debugger present
	  printf(" [Ret/Esc]\n"); // ask whether we should continue or panic

	  if (getchar() == '\033') // esc pressed?
	    goto generic;	// panic
	}

      return true;		// success -- consume int3

    nostr:
      printf("KDB: int3\n");
      goto generic;		// enter the kernel debugger
    }

  // privileged tasks also may invoke the kernel debugger with a debug
  // exception
  if (ts->trapno == 1
      && (ts->eflags & EFL_IOPL) == EFL_IOPL_USER) // only allow priv tasks
    {
      goto generic;
    }

  // can't handle trap -- kill the thread

fail:
  printf("KERNEL: %x.%x (tcb=0x%x) killed: unhandled trap\n",
	 unsigned(space_index()), id().id.lthread, (unsigned)this);
  trap_dump(ts);

  if (config::conservative)
    kdb::ke("thread killed");

  if (state_change_safely(~Thread_running, Thread_cancel|Thread_dead))
    {
      // we haven't been re-initialized (cancel was not set) -- so sleep
      while (! (state() & Thread_running))
	schedule();
    }
  
  return true;			
    
generic:
  if (!nested_trap_handler)
    return false;

  cli();
  //  kdb::com_port_init();		// re-initialize the GDB serial port

  // run the nested trap handler on a separate stack
  // equiv of: return nested_trap_handler(ts) == 0 ? true : false;

  int ret;
  static char nested_handler_stack[PAGE_SIZE];

  asm volatile
    (" movl %%esp,%%eax \n"
     " orl  %%ebx,%%ebx \n"	// don't set %esp if gdb fault recovery
     " jz   1f \n"
     " movl %%ebx,%%esp \n"
     "1: pushl %%eax \n"
     " pushl %%ecx \n"
     " call *%%edx \n"
     " addl $4,%%esp \n"
     " popl %%esp"
     : "=a" (ret)
     : "b" (gdb_trap_recover ?	// gdb fault recovery?
	    0 : (nested_handler_stack + sizeof(nested_handler_stack))),
       "c" (ts),
       "d" (nested_trap_handler)
     : "ecx", "edx", "memory");
     
  return ret == 0 ? true : false;
}

// the global page fault handler switch
inline bool
thread_t::handle_page_fault(vm_offset_t pfa, unsigned error_code)
{
  if (pfa < kmem::mem_user_max)	// page fault in user space
    {
      if (space_index() == config::sigma0_taskno)
	{
	  // special case: sigma0 can map in anything from the kernel

	  return ((cpu.feature_flags & CPUF_4MB_PAGES) 
		  ? space()->v_insert(trunc_superpage(pfa), 
				      trunc_superpage(pfa), 
				      SUPERPAGE_SIZE)
		  : space()->v_insert(trunc_page(pfa), 
				      trunc_page(pfa), 
				      PAGE_SIZE))
	    != space_t::Insert_err_nomem;
	}

      // user mode page fault -- send pager request
      return handle_page_fault_pager(pfa, error_code);
    }

  else if (error_code & 4)	// user mode?
    {
      return false;		// disallow access after mem_user_max
    }

  // we're in kernel mode

  else if ((! (error_code & 1))	// page not present
	   && kmem::virt_to_phys(reinterpret_cast<void*>(pfa)) != 0xffffffff)
    {
      // the global page directory has a page which is not present in
      // the thread's local page dir -- copy

      // This is the only path that should be executed with interrupts
      // disabled if the page faulter also had interrupts disabled.
      // thread_page_fault() takes care of that.

      space()->kmem_update(pfa);
      space()->switchin_context();
      return true;
    }

  else if ((pfa & ~(config::thread_block_size * (1 << (7 + 11)) - 1))
	   == kmem::mem_tcbs)	// in the tcb space
    {
      if (! (error_code & 1))	// page not present
	{
	  // in case of read fault, just map in the shared zero page
	  // otherwise -> allocate
	  if (!kmem::page_alloc(pfa & ~PAGE_MASK, 
				(error_code & 2) ? kmem::zero_fill
				                 : kmem::zero_map))
	    panic("can't alloc kernel page");
	}
      else
	{	  
	  // protection fault
	  // this can only be because we have the zero page mapped
	  kmem::page_free(reinterpret_cast<void*>(pfa & ~PAGE_MASK));
	  if (! kmem::page_alloc(pfa & ~PAGE_MASK, kmem::zero_fill))
	    {
	      // error could mean: someone else was faster allocating
	      // a page there, or we just don't have any pages left; verify
	      if (kmem::virt_to_phys(reinterpret_cast<void*>(pfa)) == 0xffffffff)
		panic("can't alloc kernel page");

	      // otherwise, there's a page mapped.  continue
	    }
	}

      space()->kmem_update(pfa);
      space()->switchin_context();
      return true;
    }

  else if (pfa >= kmem::ipc_window(0) 
	   && pfa < kmem::ipc_window(1) + (SUPERPAGE_SIZE<<1))
    {
      // page fault in IPC window
      return handle_ipc_page_fault(pfa);
    }
  
  printf("KERNEL: no page fault handler for region: 0x%x, error 0x%x\n",
	 pfa, error_code);

  return false;
}

bool log_page_fault = false;

// The low-level page fault handler called from entry.S.  We're
// invoked with interrupts turned off.
extern "C" int 
thread_page_fault(vm_offset_t pfa, unsigned error_code)
{
  extern unsigned gdb_trap_recover; // in OSKit's gdb_trap.c

  if (gdb_trap_recover)
    return false; // we're in the GDB stub -- let generic handler handle it

#if 1
  if (log_page_fault)
    printf("*P[%x,%x,%x] ", pfa, error_code & 0xffff, 
	   (unsigned) __builtin_return_address(1));
#endif

  // Handle the following situation with disabled interrupts: When
  // we've interrupted a context (1) in the kernel (2) with disabled
  // interrupts, (3) the page fault address is in the kernel region,
  // (4) the error code is "not mapped" (as opposed to "access
  // error"), (5) and the region is actually valid (that is, mapped in
  // kmem's shared page directory, just not in the currently active
  // page directory).
  if (! (error_code & 4))	// (1) in kernel?
    {
      unsigned *eflags;
#ifdef NO_FRAME_PTR
# error "FIXME: Missing implementation: Reading context's EFLAGS"
#else
      asm("movl (%%ebp), %0 ; leal 12(%0), %0" : "=r" (eflags));
#endif
      if (! (*eflags & EFL_IF))	// (2) interrupts were disabled
	{
	  if (pfa >= kmem::mem_user_max	// (3) page fault in kernel region
	      && (! (error_code & 1))   // (4) "page not present"
	      && kmem::virt_to_phys(reinterpret_cast<void*>(pfa)) 
	         != 0xffffffff)		// (5) but actually there
	    {
	      // OK, remain cli'd
	    }
	  else
	    {
	      // Error: We interrupted a cli'd kernel context!
	      if (! log_page_fault)
		printf("*P[%x,%x,%x] ", pfa, error_code & 0xffff, 
		       (unsigned) __builtin_return_address(1));
	      kdb::ke("page fault in cli mode");
	    }
	}
      else
	sti();			// Otherwise, enable irqs.
    }
  else
    sti();			// Otherwise, enable irqs.

  return current()->handle_page_fault(pfa, error_code);
}

// Restore a saved FPU state (or initialize the FPU).  

// Like Linux, we do not do lazy FPU saving.  Here's a comment from
// the Linux kernel on the subject: ``We fsave/fwait [in the context
// switching routine] so that an exception goes off at the right time
// (as a call from the fsave or fwait in effect) rather than to the
// wrong process.  Lazy FP saving no longer makes any sense with
// modern CPU's, and this simplifies a lot of things (SMP and UP
// become the same).''
inline void 
thread_t::handle_fpu_trap()
{
  clear_ts();

  if (state() & Thread_fpu_used)
    {
      asm volatile("frstor %0": :"m" (fpu_regs));
    }
  else
    {
      asm("fninit");
      state_add(Thread_fpu_used);
    }
  
  state_add(Thread_fpu_active);	// so that state is saved in switch_to
}

// the "FPU not available" trap entry point
extern "C" void
thread_handle_fputrap(void)
{
  current()->handle_fpu_trap();
}


// the timer interrupt is activated on every clock tick
extern "C" void
thread_timer_interrupt(void)
{
  // we're entered with disabled interrupts; that's necessary so that
  // we can't be preempted until we've acknowledged the interrupt and
  // updated the clock

  pic_disable_irq(8);
  rtcin(RTC_INTR);		// reset clock -- 1st
#if 1
  irq_ack(8);			// acknowledge interrupt
#else
  pic_ack(8);			// acknowledge interrupt
#endif
  rtcin(RTC_INTR);		// reset clock -- 2nd
  pic_enable_irq(8);

#if 0				// screen spinner for debugging purposes
  (*(unsigned char*)(kmem::phys_to_virt(0xb8000 + 16)))++;
#endif

  // lock kernel
#if 0 /* This assert() doesn't make sense if kernel_lock is eq. to cli
	 mode because the timer interrupt is entered in cli mode. */
  assert(kernel_lock.test() == false);
#endif
  kernel_lock.lock();

  // update clock and handle timeouts
  timer::update_system_clock();

  // unlock; also, re-enable interrupts
  kernel_lock.clear();

  // need to reschedule?
  if (--timeslice_ticks_left <= 0)
    current()->schedule();
}

// the slow trap entry point
extern "C" int
thread_handle_trap(trap_state *state)
{
  return current()->handle_slow_trap(state) 
    ? 0				// success -- handler consumed trap
    : -1;
}

