// -*- c++ -*-

#ifndef TIME_H
#define TIME_H

#include "config.h"
#include "kmem.h"
#include "config.h"
#include "thread.h"
#include "thread_list.h"

class timeout_t
{
public:
  bool is_set();
  bool has_hit();

  void set(unsigned long long abs_microsec);  
  void reset();

  thread_t *owner();

  timeout_t();			// constructor

private:
  timeout_t(const timeout_t&);	// default copy constructor is undefined

  bool check_hit();

  // implementation details follow
  unsigned long long _wakeup;	// absolute time we want to be woken up
  timeout_t *_next, *_prev;	// next/prev in timer queue

  union {
    struct {
      bool set : 1;
      bool hit : 1;
    } _flags;
    unsigned short _flagword;
  };

  friend class timer;
};

class timer
{
public:
  static void update_system_clock();


private:
  timer();			// default constructors are undefined
  timer(const timer&);

  friend class timeout_t;	// timeouts can add themselves to the timeout
				// queue

  static void do_timeouts();

  // implementation details follow

  static timeout_t *first_timeout;
};

//
// inlines
//

// When profiling, do not inline, and only define functions if
// TIMEOUT_IMPLEMENTATION has been set.  OTOH when not profiling,
// always define inline implementations
#ifdef PROFILE
# define __INLINE 
#else  // !PROFILE
# define __INLINE inline
# ifndef TIMEOUT_IMPLEMENTATION
#  define TIMEOUT_IMPLEMENTATION
# endif// !TIMEOUT_IMPLEMENTATION
#endif // !PROFILE

// timeout_t
#ifdef TIMEOUT_IMPLEMENTATION

__INLINE 
timeout_t::timeout_t()
{
  _flagword = 0;
}

__INLINE bool 
timeout_t::is_set()
{
  return _flags.set;
}

__INLINE bool 
timeout_t::has_hit()
{
  return _flags.hit;
}

__INLINE thread_t *
timeout_t::owner()
{
  return reinterpret_cast<thread_t *>(reinterpret_cast<vm_offset_t>(this) 
				 & ~(config::thread_block_size - 1));
}

__INLINE void
timeout_t::set(unsigned long long abs_microsec)
{
  // XXX uses global kernel lock
  
  _wakeup = abs_microsec;

  bool was_locked = kernel_lock.test_and_set();

  _flags.set = 1;
  if (!timer::first_timeout)
    {
      timer::first_timeout = this;
      _prev = _next = 0;
    }
  else
    {
      timeout_t *ti = timer::first_timeout;

      for (;;)
	{
	  if (ti->_wakeup >= _wakeup)
	    {
	      // insert before ti
	      _next = ti;
	      _prev = ti->_prev;
	      ti->_prev = this;
	      if (_prev)
		_prev->_next = this;
	      else
		timer::first_timeout = this;
	      
	      goto out;
	    }

	  if (! ti->_next)
	    break;

	  ti = ti->_next;
	}

      // insert as last item after ti
      ti->_next = this;
      _prev = ti;
      _next = 0;
    }

 out:
  kernel_lock.set(was_locked);
}

__INLINE void 
timeout_t::reset()
{
  // XXX uses global kernel lock

  if (! is_set())
    return;			// avoid lock overhead if not set

  bool was_locked = kernel_lock.test_and_set();

  if (is_set())
    {
      if (_prev)
	_prev->_next = _next;
      else
	timer::first_timeout = _next;
      if (_next)
	_next->_prev = _prev;
    }

  _flags.set = 0;

  kernel_lock.set(was_locked);
}

__INLINE bool 
timeout_t::check_hit()
{
  // XXX assume we run kernel-locked

  if (kmem::info()->clock < _wakeup)
    return false;

  timer::first_timeout = _next;		// dequeue
  if (_next) _next->_prev = 0;
  _flags.hit = 1;
  _flags.set = 0;
  owner()->state_change(~Thread_ipc_in_progress, 
			Thread_running); // set thread running

  // XXX don't we want to switch immediately if the newly woken-up
  // thread has a higher prio?

  owner()->ready_enqueue();

  return true;
}

// timer

__INLINE void 
timer::do_timeouts()
{
  while (first_timeout)
    {
      if (! first_timeout->check_hit())
	break;
    }
}

__INLINE void
timer::update_system_clock()
{
  kmem::info()->clock += config::microsec_per_tick;

  do_timeouts();
}

#endif // TIMEOUT_IMPLEMENTATION

#undef __INLINE

//
// helpers
//

inline unsigned long long
timeout_to_microsec(unsigned man, unsigned exp)
{
  return static_cast<unsigned long long>(man) << ((15 - exp) << 1);
}

#endif
